@@ -1,9 +1,9 @@
 /*
- * "$Id: html.cxx,v 1.17.2.22 2001/12/17 18:27:33 mike Exp $"
+ * "$Id: html.cxx,v 1.17.2.28 2002/10/02 18:53:05 mike Exp $"
  *
  *   HTML exporting functions for HTMLDOC, a HTML document processing program.
  *
- *   Copyright 1997-2001 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -112,13 +112,15 @@ html_export(tree_t *document,	/* I - Document to export */
 
   if (OutputFiles && TitleImage[0] != '\0' && TitlePage &&
 #ifdef WIN32
-      stricmp(file_extension(TitleImage), "htm") != 0 &&
-      stricmp(file_extension(TitleImage), "html") != 0 &&
-      stricmp(file_extension(TitleImage), "shtml") != 0)
+      stricmp(file_extension(TitleImage), "bmp") == 0 ||
+      stricmp(file_extension(TitleImage), "gif") == 0 ||
+      stricmp(file_extension(TitleImage), "jpg") == 0 ||
+      stricmp(file_extension(TitleImage), "png") == 0)
 #else
-      strcmp(file_extension(TitleImage), "htm") != 0 &&
-      strcmp(file_extension(TitleImage), "html") != 0 &&
-      strcmp(file_extension(TitleImage), "shtml") != 0)
+      strcmp(file_extension(TitleImage), "bmp") == 0 ||
+      strcmp(file_extension(TitleImage), "gif") == 0 ||
+      strcmp(file_extension(TitleImage), "jpg") == 0 ||
+      strcmp(file_extension(TitleImage), "png") == 0)
 #endif // WIN32
     image_copy(TitleImage, OutputPath);
 
@@ -152,7 +154,8 @@ html_export(tree_t *document,	/* I - Document to export */
   {
     write_header(&out, (uchar *)"index.html", title, author, copyright,
                  docnumber, NULL);
-    write_title(out, title, author, copyright, docnumber);
+    if (out != NULL)
+      write_title(out, title, author, copyright, docnumber);
 
     write_footer(&out, NULL);
     write_header(&out, (uchar *)"toc.html", title, author, copyright,
@@ -162,7 +165,8 @@ html_export(tree_t *document,	/* I - Document to export */
     write_header(&out, (uchar *)"index.html", title, author, copyright,
                  docnumber, NULL);
 
-  write_all(out, toc, 0);
+  if (out != NULL)
+    write_all(out, toc, 0);
   write_footer(&out, NULL);
 
  /*
@@ -173,13 +177,14 @@ html_export(tree_t *document,	/* I - Document to export */
   {
     write_header(&out, htmlGetVariable(document, (uchar *)"FILENAME"),
                  title, author, copyright, docnumber, document);
-    write_all(out, document->child, 0);
+    if (out != NULL)
+      write_all(out, document->child, 0);
     write_footer(&out, document);
 
     document = document->next;
   }
 
-  if (!OutputFiles && out != stdout)
+  if (!OutputFiles && out != stdout && out != NULL)
   {
     fputs("</BODY>\n", out);
     fputs("</HTML>\n", out);
@@ -201,7 +206,7 @@ html_export(tree_t *document,	/* I - Document to export */
     links       = NULL;
   }
 
-  return (0);
+  return (out == NULL);
 }
 
 
@@ -274,7 +279,8 @@ write_header(FILE   **out,	/* IO - Output file */
           "\"http://www.w3.org/TR/REC-html40/loose.dtd\">\n", *out);
     fputs("<HTML>\n", *out);
     fputs("<HEAD>\n", *out);
-    fprintf(*out, "<TITLE>%s</TITLE>\n", title);
+    if (title != NULL)
+      fprintf(*out, "<TITLE>%s</TITLE>\n", title);
     if (author != NULL)
       fprintf(*out, "<META NAME=\"author\" CONTENT=\"%s\">\n", author);
     if (copyright != NULL)
@@ -402,6 +408,7 @@ write_title(FILE  *out,		/* I - Output file */
             uchar *docnumber)	/* I - ID number for document */
 {
   FILE		*fp;		/* Title file */
+  const char	*title_file;	/* Location of title file */
   tree_t	*t;		/* Title file document tree */
 
 
@@ -409,17 +416,29 @@ write_title(FILE  *out,		/* I - Output file */
     return;
 
 #ifdef WIN32
-  if (stricmp(file_extension(TitleImage), "htm") == 0 ||
-      stricmp(file_extension(TitleImage), "html") == 0 ||
-      stricmp(file_extension(TitleImage), "shtml") == 0)
+  if (TitleImage[0] &&
+      stricmp(file_extension(TitleImage), "bmp") != 0 &&
+      stricmp(file_extension(TitleImage), "gif") != 0 &&
+      stricmp(file_extension(TitleImage), "jpg") != 0 &&
+      stricmp(file_extension(TitleImage), "png") != 0)
 #else
-  if (strcmp(file_extension(TitleImage), "htm") == 0 ||
-      strcmp(file_extension(TitleImage), "html") == 0 ||
-      strcmp(file_extension(TitleImage), "shtml") == 0)
+  if (TitleImage[0] &&
+      strcmp(file_extension(TitleImage), "bmp") != 0 &&
+      strcmp(file_extension(TitleImage), "gif") != 0 &&
+      strcmp(file_extension(TitleImage), "jpg") != 0 &&
+      strcmp(file_extension(TitleImage), "png") != 0)
 #endif // WIN32
   {
+    // Find the title page file...
+    if ((title_file = file_find(Path, TitleImage)) == NULL)
+    {
+      progress_error(HD_ERROR_FILE_NOT_FOUND,
+                     "Unable to find title file \"%s\"!", TitleImage);
+      return;
+    }
+
     // Write a title page from HTML source...
-    if ((fp = fopen(TitleImage, "rb")) == NULL)
+    if ((fp = fopen(title_file, "rb")) == NULL)
     {
       progress_error(HD_ERROR_FILE_NOT_FOUND,
                      "Unable to open title file \"%s\" - %s!",
@@ -530,6 +549,7 @@ write_all(FILE   *out,		/* I - Output file */
 	  break;
 
       case MARKUP_COMMENT :
+      case MARKUP_UNKNOWN :
           fprintf(out, "\n<!--%s-->\n", t->data);
 	  break;
 
@@ -556,6 +576,15 @@ write_all(FILE   *out,		/* I - Output file */
       case MARKUP_H4 :
       case MARKUP_H5 :
       case MARKUP_H6 :
+      case MARKUP_H7 :
+      case MARKUP_H8 :
+      case MARKUP_H9 :
+      case MARKUP_H10 :
+      case MARKUP_H11 :
+      case MARKUP_H12 :
+      case MARKUP_H13 :
+      case MARKUP_H14 :
+      case MARKUP_H15 :
       case MARKUP_HR :
       case MARKUP_LI :
       case MARKUP_OL :
@@ -601,6 +630,9 @@ write_all(FILE   *out,		/* I - Output file */
 	          t->markup == MARKUP_IMG)
 		continue;
 
+              if (strncasecmp((char *)t->vars[i].name, "_HD_", 4) == 0)
+	        continue;
+
 	      if (col > 72 && !t->preformatted)
 	      {
         	putc('\n', out);
@@ -668,6 +700,7 @@ write_all(FILE   *out,		/* I - Output file */
 	case MARKUP_NOBR :
 	case MARKUP_SPACER :
 	case MARKUP_WBR :
+	case MARKUP_UNKNOWN :
             break;
 
         case MARKUP_CENTER :
@@ -680,6 +713,15 @@ write_all(FILE   *out,		/* I - Output file */
         case MARKUP_H4 :
         case MARKUP_H5 :
         case MARKUP_H6 :
+	case MARKUP_H7 :
+	case MARKUP_H8 :
+	case MARKUP_H9 :
+	case MARKUP_H10 :
+	case MARKUP_H11 :
+	case MARKUP_H12 :
+	case MARKUP_H13 :
+	case MARKUP_H14 :
+	case MARKUP_H15 :
         case MARKUP_LI :
         case MARKUP_OL :
         case MARKUP_P :
@@ -942,5 +984,5 @@ update_links(tree_t *t,		/* I - Document tree */
 
 
 /*
- * End of "$Id: html.cxx,v 1.17.2.22 2001/12/17 18:27:33 mike Exp $".
+ * End of "$Id: html.cxx,v 1.17.2.28 2002/10/02 18:53:05 mike Exp $".
  */
