@@ -1,9 +1,9 @@
 /*
- * "$Id: html.h,v 1.9.2.6 2001/10/31 17:15:54 mike Exp $"
+ * "$Id: html.h,v 1.9.2.10 2002/07/27 03:41:31 mike Exp $"
  *
  *   HTML parsing definitions for HTMLDOC, a HTML document processing program.
  *
- *   Copyright 1997-2001 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -57,8 +57,9 @@ extern "C" {
 
 typedef enum
 {
-	MARKUP_FILE = -2,	/* File Delimiter */
-	MARKUP_ERROR = -1,
+	MARKUP_FILE = -3,	/* File Delimiter */
+	MARKUP_UNKNOWN = -2,	/* Unknown element */
+	MARKUP_ERROR = -1,	
 	MARKUP_NONE = 0,
 	MARKUP_COMMENT,
 	MARKUP_DOCTYPE,
@@ -100,6 +101,15 @@ typedef enum
 	MARKUP_H4,
 	MARKUP_H5,
 	MARKUP_H6,
+	MARKUP_H7,
+	MARKUP_H8,
+	MARKUP_H9,
+	MARKUP_H10,
+	MARKUP_H11,
+	MARKUP_H12,
+	MARKUP_H13,
+	MARKUP_H14,
+	MARKUP_H15,
 	MARKUP_HEAD,
 	MARKUP_HR,
 	MARKUP_HTML,
@@ -301,6 +311,8 @@ extern void	htmlSetBaseSize(float p, float s);
 extern void	htmlSetCharSet(const char *cs);
 extern void	htmlSetTextColor(uchar *color);
 
+extern void	htmlDebugStats(const char *title, tree_t *t);
+
 #  ifdef __cplusplus
 }
 #  endif /* __cplusplus */
@@ -308,5 +320,5 @@ extern void	htmlSetTextColor(uchar *color);
 #endif /* !_HTML_H_ */
 
 /*
- * End of "$Id: html.h,v 1.9.2.6 2001/10/31 17:15:54 mike Exp $".
+ * End of "$Id: html.h,v 1.9.2.10 2002/07/27 03:41:31 mike Exp $".
  */
