@@ -1,9 +1,9 @@
 /*
- * "$Id: htmldoc.cxx,v 1.36.2.35 2001/12/17 18:27:33 mike Exp $"
+ * "$Id: htmldoc.cxx,v 1.36.2.55 2002/10/25 13:37:54 mike Exp $"
  *
  *   Main entry for HTMLDOC, a HTML document processing program.
  *
- *   Copyright 1997-2001 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -78,7 +78,8 @@ static int	compare_strings(const char *s, const char *t, int tmin);
 static int	load_book(const char *filename, tree_t **document,
 		          exportfunc_t *exportfunc);
 static void	parse_options(const char *line, exportfunc_t *exportfunc);
-static int	read_file(const char *filename, tree_t **document);
+static int	read_file(const char *filename, tree_t **document,
+		          const char *path);
 static void	term_handler(int signum);
 static void	usage(void);
 
@@ -111,11 +112,22 @@ main(int  argc,		/* I - Number of command-line arguments */
   const char	*extension;	/* Extension of output filename */
   float		fontsize,	/* Base font size */
 		fontspacing;	/* Base font spacing */
-#ifdef HAVE_LIBFLTK
-  int		display_set;	/* True if display set */
-#endif // HAVE_LIBFLTK
+  int		num_files;	/* Number of files provided on the command-line */
 
 
+#ifdef __APPLE__
+ /*
+  * OSX passes an extra command-line option when run from the Finder.
+  * If the first command-line argument is "-psn..." then skip it...
+  */
+
+  if (argc > 1 && strncmp(argv[1], "-psn", 4) == 0)
+  {
+    argv ++;
+    argc --;
+  }
+#endif // __APPLE__
+
  /*
   * Localize as needed...
   */
@@ -154,16 +166,17 @@ main(int  argc,		/* I - Number of command-line arguments */
 
   fontsize    = 11.0f;
   fontspacing = 1.2f;
-#ifdef HAVE_LIBFLTK
-  display_set = 1;
-#endif // HAVE_LIBFLTK
+  num_files   = 0;
 
   for (i = 1; i < argc; i ++)
     if (compare_strings(argv[i], "--batch", 4) == 0)
     {
       i ++;
       if (i < argc)
+      {
+        num_files ++;
         load_book(argv[i], &document, &exportfunc);
+      }
       else
         usage();
     }
@@ -217,7 +230,16 @@ main(int  argc,		/* I - Number of command-line arguments */
     {
       i ++;
       if (i < argc)
+      {
         _htmlBrowserWidth = atof(argv[i]);
+
+	if (_htmlBrowserWidth < 1.0f)
+	{
+	  progress_error(HD_ERROR_INTERNAL_ERROR, "Bad browser width \"%s\"!",
+	                 argv[i]);
+	  usage();
+	}
+      }
       else
         usage();
     }
@@ -237,9 +259,9 @@ main(int  argc,		/* I - Number of command-line arguments */
     else if (compare_strings(argv[i], "--compression", 5) == 0 ||
              strncmp(argv[i], "--compression=", 14) == 0)
     {
-      if (strlen(argv[i]) > 14 && PDFVersion >= 1.2)
+      if (strlen(argv[i]) > 14 && PDFVersion >= 12)
         Compression = atoi(argv[i] + 14);
-      else if (PDFVersion >= 1.2)
+      else if (PDFVersion >= 12)
         Compression = 1;
     }
     else if (compare_strings(argv[i], "--continuous", 5) == 0)
@@ -265,7 +287,6 @@ main(int  argc,		/* I - Number of command-line arguments */
       // The X standard requires support for the -display option, but
       // we also support the GNU standard --display...
       i ++;
-      display_set = 3;
       if (i < argc)
         Fl::display(argv[i]);
       else
@@ -278,10 +299,21 @@ main(int  argc,		/* I - Number of command-line arguments */
     {
       i ++;
       if (i < argc)
+      {
         PDFEffectDuration = atof(argv[i]);
+
+	if (PDFEffectDuration < 0.0f)
+	{
+	  progress_error(HD_ERROR_INTERNAL_ERROR, "Bad effect duration \"%s\"!",
+	                 argv[i]);
+	  usage();
+	}
+      }
       else
         usage();
     }
+    else if (compare_strings(argv[i], "--embedfonts", 4) == 0)
+      EmbedFonts = 1;
     else if (compare_strings(argv[i], "--encryption", 4) == 0)
       Encryption = 1;
     else if (compare_strings(argv[i], "--firstpage", 4) == 0)
@@ -365,26 +397,26 @@ main(int  argc,		/* I - Number of command-line arguments */
 	{
           exportfunc = (exportfunc_t)pspdf_export;
 	  PSLevel    = 0;
-	  PDFVersion = 1.4;
+	  PDFVersion = 14;
 	}
         else if (strcasecmp(argv[i], "pdf13") == 0 ||
 	         strcasecmp(argv[i], "pdf") == 0)
 	{
           exportfunc = (exportfunc_t)pspdf_export;
 	  PSLevel    = 0;
-	  PDFVersion = 1.3;
+	  PDFVersion = 13;
 	}
         else if (strcasecmp(argv[i], "pdf12") == 0)
 	{
           exportfunc = (exportfunc_t)pspdf_export;
 	  PSLevel    = 0;
-	  PDFVersion = 1.2;
+	  PDFVersion = 12;
 	}
         else if (strcasecmp(argv[i], "pdf11") == 0)
 	{
           exportfunc  = (exportfunc_t)pspdf_export;
 	  PSLevel     = 0;
-	  PDFVersion  = 1.1;
+	  PDFVersion  = 11;
 	  Compression = 0;
 	}
         else if (strcasecmp(argv[i], "html") == 0)
@@ -577,7 +609,9 @@ main(int  argc,		/* I - Number of command-line arguments */
       Compression = 0;
     else if (compare_strings(argv[i], "--no-duplex", 4) == 0)
       PageDuplex = 0;
-    else if (compare_strings(argv[i], "--no-encryption", 6) == 0)
+    else if (compare_strings(argv[i], "--no-embedfonts", 7) == 0)
+      EmbedFonts = 0;
+    else if (compare_strings(argv[i], "--no-encryption", 7) == 0)
       Encryption = 0;
     else if (compare_strings(argv[i], "--no-jpeg", 6) == 0)
       OutputJPEG = 0;
@@ -589,16 +623,33 @@ main(int  argc,		/* I - Number of command-line arguments */
       TocNumbers = 0;
     else if (compare_strings(argv[i], "--no-pscommands", 6) == 0)
       PSCommands = 0;
+    else if (compare_strings(argv[i], "--no-strict", 6) == 0)
+      StrictHTML = 0;
     else if (compare_strings(argv[i], "--no-title", 7) == 0)
       TitlePage = 0;
     else if (compare_strings(argv[i], "--no-toc", 7) == 0)
       TocLevels = 0;
     else if (compare_strings(argv[i], "--no-truetype", 7) == 0)
-      TrueType = 0;
+    {
+      fputs("htmldoc: Warning, --no-truetype option superceded by --no-embedfonts!\n", stderr);
+      EmbedFonts = 0;
+    }
     else if (compare_strings(argv[i], "--no-xrxcomments", 6) == 0)
       XRXComments = 0;
-    else if (compare_strings(argv[i], "--numbered", 4) == 0)
+    else if (compare_strings(argv[i], "--numbered", 5) == 0)
       TocNumbers = 1;
+    else if (compare_strings(argv[i], "--nup", 5) == 0)
+    {
+      i ++;
+      if (i >= argc)
+        usage();
+
+      NumberUp = atoi(argv[i]);
+
+      if (NumberUp != 1 && NumberUp != 2 && NumberUp != 4 &&
+          NumberUp != 6 && NumberUp != 9 && NumberUp != 16)
+	usage();
+    }
     else if (compare_strings(argv[i], "--outdir", 6) == 0 ||
              strcmp(argv[i], "-d") == 0)
     {
@@ -656,7 +707,16 @@ main(int  argc,		/* I - Number of command-line arguments */
     {
       i ++;
       if (i < argc)
+      {
         PDFPageDuration = atof(argv[i]);
+
+	if (PDFPageDuration < 1.0f)
+	{
+	  progress_error(HD_ERROR_INTERNAL_ERROR, "Bad page duration \"%s\"!",
+	                 argv[i]);
+	  usage();
+	}
+      }
       else
         usage();
     }
@@ -736,6 +796,9 @@ main(int  argc,		/* I - Number of command-line arguments */
         Permissions |= PDF_PERM_ANNOTATE;
       else if (strcasecmp(argv[i], "no-annotate") == 0)
         Permissions &= ~PDF_PERM_ANNOTATE;
+
+      if (Permissions != -4)
+        Encryption = 1;
     }
     else if (compare_strings(argv[i], "--portrait", 4) == 0)
       Landscape = 0;
@@ -763,7 +826,7 @@ main(int  argc,		/* I - Number of command-line arguments */
       else
         usage();
     }
-    else if (compare_strings(argv[i], "--size", 3) == 0)
+    else if (compare_strings(argv[i], "--size", 4) == 0)
     {
       i ++;
       if (i < argc)
@@ -771,6 +834,8 @@ main(int  argc,		/* I - Number of command-line arguments */
       else
         usage();
     }
+    else if (compare_strings(argv[i], "--strict", 4) == 0)
+      StrictHTML = 1;
     else if (compare_strings(argv[i], "--textcolor", 7) == 0)
     {
       i ++;
@@ -847,12 +912,21 @@ main(int  argc,		/* I - Number of command-line arguments */
         usage();
     }
     else if (compare_strings(argv[i], "--truetype", 4) == 0)
-      TrueType = 1;
-    else if (compare_strings(argv[i], "--verbose", 3) == 0 ||
+    {
+      fputs("htmldoc: Warning, --truetype option superceded by --embedfonts!\n", stderr);
+
+      EmbedFonts = 1;
+    }
+    else if (compare_strings(argv[i], "--verbose", 6) == 0 ||
              strcmp(argv[i], "-v") == 0)
     {
       Verbosity ++;
     }
+    else if (compare_strings(argv[i], "--version", 6) == 0)
+    {
+      puts(SVERSION);
+      return (0);
+    }
     else if (compare_strings(argv[i], "--webpage", 3) == 0)
     {
       TocLevels    = 0;
@@ -869,7 +943,9 @@ main(int  argc,		/* I - Number of command-line arguments */
       * Read from stdin...
       */
 
-      _htmlPPI = _htmlBrowserWidth / PageWidth * 72.0f;
+      num_files ++;
+
+      _htmlPPI = 72.0f * _htmlBrowserWidth / (PageWidth - PageLeft - PageRight);
 
       file = htmlAddTree(NULL, MARKUP_FILE, NULL);
       htmlSetVariable(file, (uchar *)"FILENAME", (uchar *)"");
@@ -884,23 +960,18 @@ main(int  argc,		/* I - Number of command-line arguments */
       _fsetmode(stdin, "b");
 #endif // WIN32 || __EMX__
 
-      htmlReadFile(file, stdin, "");
+      htmlReadFile(file, stdin, ".");
 
-      if (file->child != NULL)
-      {
-        if (document == NULL)
-          document = file;
-        else
-        {
-          while (document->next != NULL)
-            document = document->next;
-
-          document->next = file;
-          file->prev     = document;
-        }
-      }
+      if (document == NULL)
+        document = file;
       else
-        htmlDeleteTree(file);
+      {
+        while (document->next != NULL)
+          document = document->next;
+
+        document->next = file;
+        file->prev     = document;
+      }
     }
     else if (argv[i][0] == '-')
       usage();
@@ -919,20 +990,19 @@ main(int  argc,		/* I - Number of command-line arguments */
     }
 #endif /* HAVE_LIBFLTK */
     else
-      read_file(argv[i], &document);
+    {
+      num_files ++;
+
+      read_file(argv[i], &document, Path);
+    }
 
  /*
   * Display the GUI if necessary...
   */
 
 #ifdef HAVE_LIBFLTK
-  if (document == NULL && BookGUI == NULL)
-  {
-    if (argc == display_set)
-      BookGUI = new GUI();
-    else
-      usage();
-  }
+  if (num_files == 0 && BookGUI == NULL)
+    BookGUI = new GUI();
 
   if (BookGUI != NULL)
   {
@@ -952,7 +1022,7 @@ main(int  argc,		/* I - Number of command-line arguments */
   * We *must* have a document to process...
   */
 
-  if (document == NULL)
+  if (num_files == 0 || document == NULL)
   {
     puts("ERROR: No HTML files!");
     usage();
@@ -965,6 +1035,8 @@ main(int  argc,		/* I - Number of command-line arguments */
   while (document->prev != NULL)
     document = document->prev;
 
+  htmlDebugStats("Document Tree", document);
+
  /*
   * Build a table of contents for the documents if necessary...
   */
@@ -974,6 +1046,8 @@ main(int  argc,		/* I - Number of command-line arguments */
   else
     toc = NULL;
 
+  htmlDebugStats("Table of Contents Tree", toc);
+
  /*
   * Generate the output file(s).
   */
@@ -1158,6 +1232,8 @@ prefs_load(void)
 	get_format(line + 11, Header);
       else if (strncasecmp(line, "PAGEFOOTER=", 11) == 0)
 	get_format(line + 11, Footer);
+      else if (strncasecmp(line, "NUMBERUP=", 9) == 0)
+        NumberUp = atoi(line + 9);
       else if (strncasecmp(line, "TOCHEADER=", 10) == 0)
 	get_format(line + 10, TocHeader);
       else if (strncasecmp(line, "TOCFOOTER=", 10) == 0)
@@ -1180,7 +1256,12 @@ prefs_load(void)
       else if (strncasecmp(line, "HEADFOOTSIZE=", 13) == 0)
 	HeadFootSize = atof(line + 13);
       else if (strncasecmp(line, "PDFVERSION=", 11) == 0)
-	PDFVersion = atof(line + 11);
+      {
+        if (strchr(line + 11, '.') != NULL)
+	  PDFVersion = (int)(atof(line + 11) * 10.0 + 0.5);
+	else
+	  PDFVersion = atoi(line + 11);
+      }
       else if (strncasecmp(line, "PSLEVEL=", 8) == 0)
 	PSLevel = atoi(line + 8);
       else if (strncasecmp(line, "PSCOMMANDS=", 11) == 0)
@@ -1218,7 +1299,9 @@ prefs_load(void)
       else if (strncasecmp(line, "LINKS=", 6) == 0)
         Links = atoi(line + 6);
       else if (strncasecmp(line, "TRUETYPE=", 9) == 0)
-        TrueType = atoi(line + 9);
+        EmbedFonts = atoi(line + 9);
+      else if (strncasecmp(line, "EMBEDFONTS=", 11) == 0)
+        EmbedFonts = atoi(line + 11);
       else if (strncasecmp(line, "PATH=", 5) == 0)
       {
 	strncpy(Path, line + 5, sizeof(Path) - 1);
@@ -1229,6 +1312,9 @@ prefs_load(void)
 	strncpy(Proxy, line + 6, sizeof(Proxy) - 1);
 	Proxy[sizeof(Proxy) - 1] = '\0';
       }
+      else if (strncasecmp(line, "STRICTHTML=", 11) == 0)
+        StrictHTML = atoi(line + 11);
+
 #  ifdef HAVE_LIBFLTK
       else if (strncasecmp(line, "EDITOR=", 7) == 0)
         strcpy(HTMLEditor, line + 7);
@@ -1308,6 +1394,7 @@ prefs_save(void)
     fprintf(fp, "JPEG=%d\n", OutputJPEG);
     fprintf(fp, "PAGEHEADER=%s\n", get_fmt(Header));
     fprintf(fp, "PAGEFOOTER=%s\n", get_fmt(Footer));
+    fprintf(fp, "NUMBERUP=%d\n", NumberUp);
     fprintf(fp, "TOCHEADER=%s\n", get_fmt(TocHeader));
     fprintf(fp, "TOCFOOTER=%s\n", get_fmt(TocFooter));
     fprintf(fp, "TOCTITLE=%s\n", TocTitle);
@@ -1319,7 +1406,7 @@ prefs_save(void)
     fprintf(fp, "HEADFOOTTYPE=%d\n", HeadFootType);
     fprintf(fp, "HEADFOOTSTYLE=%d\n", HeadFootStyle);
     fprintf(fp, "HEADFOOTSIZE=%.2f\n", HeadFootSize);
-    fprintf(fp, "PDFVERSION=%.1f\n", PDFVersion);
+    fprintf(fp, "PDFVERSION=%d\n", PDFVersion);
     fprintf(fp, "PSLEVEL=%d\n", PSLevel);
     fprintf(fp, "PSCOMMANDS=%d\n", PSCommands);
     fprintf(fp, "XRXCOMMENTS=%d\n", XRXComments);
@@ -1335,9 +1422,10 @@ prefs_save(void)
     fprintf(fp, "OWNERPASSWORD=%s\n", OwnerPassword);
     fprintf(fp, "USERPASSWORD=%s\n", UserPassword);
     fprintf(fp, "LINKS=%d\n", Links);
-    fprintf(fp, "TRUETYPE=%d\n", TrueType);
+    fprintf(fp, "EMBEDFONTS=%d\n", EmbedFonts);
     fprintf(fp, "PATH=%s\n", Path);
     fprintf(fp, "PROXY=%s\n", Proxy);
+    fprintf(fp, "STRICTHTML=%d\n", StrictHTML);
 
 #ifdef HAVE_LIBFLTK
     fprintf(fp, "EDITOR=%s\n", HTMLEditor);
@@ -1382,25 +1470,28 @@ load_book(const char   *filename,	// I  - Book file
   FILE		*fp;			// File to read from
   char		line[10240];		// Line from file
   const char 	*dir;			// Directory
+  const char	*local;			// Local filename
+  char		path[2048];		// Current path
 
 
-  // If the filename contains a path, chdir to it first...
-  if ((dir = file_directory(filename)) != NULL)
+  // See if the filename contains a path...
+  dir = file_directory(filename);
+
+  if (dir != NULL)
+    snprintf(path, sizeof(path), "%s;%s", dir, Path);
+  else
   {
-   /*
-    * Filename contains a complete path - get the directory portion and do
-    * a chdir()...
-    */
-
-    chdir(dir);
-
-    filename = file_basename(filename);
+    strncpy(path, Path, sizeof(path) - 1);
+    path[sizeof(path) - 1] = '\0';
   }
 
   // Open the file...
-  if ((fp = fopen(filename, "rb")) == NULL)
+  if ((local = file_find(Path, filename)) == NULL)
+    return (0);
+
+  if ((fp = fopen(local, "rb")) == NULL)
   {
-    fprintf(stderr, "htmldoc: Unable to open \"%s\": %s\n", filename,
+    fprintf(stderr, "htmldoc: Unable to open \"%s\": %s\n", local,
             strerror(errno));
     return (0);
   }
@@ -1410,7 +1501,8 @@ load_book(const char   *filename,	// I  - Book file
   if (strncmp(line, "#HTMLDOC", 8) != 0)
   {
     fclose(fp);
-    fprintf(stderr, "htmldoc: Bad or missing #HTMLDOC header in %s.", filename);
+    progress_error(HD_ERROR_BAD_FORMAT,
+                   "htmldoc: Bad or missing #HTMLDOC header in %s.", filename);
     return (0);
   }
 
@@ -1421,7 +1513,17 @@ load_book(const char   *filename,	// I  - Book file
     file_gets(line, sizeof(line), fp);
 
     if (line[0] == '-')
+    {
       parse_options(line, exportfunc);
+
+      if (dir != NULL)
+	snprintf(path, sizeof(path), "%s;%s", dir, Path);
+      else
+      {
+	strncpy(path, Path, sizeof(path) - 1);
+	path[sizeof(path) - 1] = '\0';
+      }
+    }
   }
   while (!line[0]);			// Skip blank lines
 
@@ -1431,11 +1533,21 @@ load_book(const char   *filename,	// I  - Book file
     if (!line[0])
       continue;				// Skip blank lines
     else if (line[0] == '-')
+    {
       parse_options(line, exportfunc);
+
+      if (dir != NULL)
+	snprintf(path, sizeof(path), "%s;%s", dir, Path);
+      else
+      {
+	strncpy(path, Path, sizeof(path) - 1);
+	path[sizeof(path) - 1] = '\0';
+      }
+    }
     else if (line[0] == '\\')
-      read_file(line + 1, document);
+      read_file(line + 1, document, path);
     else
-      read_file(line, document);
+      read_file(line, document, path);
   }
 
   // Close the book file and return...
@@ -1520,14 +1632,16 @@ parse_options(const char   *line,	// I - Options from book file
       Links = 0;
       continue;
     }
-    else if (strcmp(temp, "--truetype") == 0)
+    else if (strcmp(temp, "--embedfonts") == 0 ||
+             strcmp(temp, "--truetype") == 0)
     {
-      TrueType = 1;
+      EmbedFonts = 1;
       continue;
     }
-    else if (strcmp(temp, "--no-truetype") == 0)
+    else if (strcmp(temp, "--no-embedfonts") == 0 ||
+             strcmp(temp, "--no-truetype") == 0)
     {
-      TrueType = 0;
+      EmbedFonts = 0;
       continue;
     }
     else if (strcmp(temp, "--pscommands") == 0)
@@ -1664,26 +1778,26 @@ parse_options(const char   *line,	// I - Options from book file
       {
         *exportfunc = (exportfunc_t)pspdf_export;
 	PSLevel     = 0;
-	PDFVersion  = 1.1f;
+	PDFVersion  = 11;
       }
       else if (strcmp(temp2, "pdf12") == 0)
       {
         *exportfunc = (exportfunc_t)pspdf_export;
 	PSLevel     = 0;
-	PDFVersion  = 1.2f;
+	PDFVersion  = 12;
       }
       else if (strcmp(temp2, "pdf") == 0 ||
                strcmp(temp2, "pdf13") == 0)
       {
         *exportfunc = (exportfunc_t)pspdf_export;
 	PSLevel     = 0;
-	PDFVersion  = 1.3f;
+	PDFVersion  = 13;
       }
       else if (strcmp(temp2, "pdf14") == 0)
       {
         *exportfunc = (exportfunc_t)pspdf_export;
 	PSLevel     = 0;
-	PDFVersion  = 1.4f;
+	PDFVersion  = 14;
       }
     }
     else if (strcmp(temp, "--logo") == 0 ||
@@ -1712,6 +1826,8 @@ parse_options(const char   *line,	// I - Options from book file
     }
     else if (strcmp(temp, "--browserwidth") == 0)
       _htmlBrowserWidth = atof(temp2);
+    else if (strcmp(temp, "--nup") == 0)
+      NumberUp = atoi(temp2);
     else if (strcmp(temp, "--size") == 0)
       set_page_size(temp2);
     else if (strcmp(temp, "--left") == 0)
@@ -1973,7 +2089,8 @@ parse_options(const char   *line,	// I - Options from book file
 
 static int				// O  - 1 on success, 0 on failure
 read_file(const char *filename,		// I  - File/URL to read
-          tree_t     **document)	// IO - Current document
+          tree_t     **document,	// IO - Current document
+	  const char *path)		// I  - Search path
 {
   FILE		*docfile;		// Document file
   tree_t	*file;			// HTML document file
@@ -1981,7 +2098,10 @@ read_file(const char *filename,		// I  - File/URL to read
   char		base[1024];		// Base directory name of file
 
 
-  if ((realname = file_find(Path, filename)) != NULL)
+  DEBUG_printf(("read_file(filename=\"%s\", document=%p, path=\"%s\")\n",
+                filename, document, path));
+
+  if ((realname = file_find(path, filename)) != NULL)
   {
     if ((docfile = fopen(realname, "rb")) != NULL)
     {
@@ -1992,7 +2112,7 @@ read_file(const char *filename,		// I  - File/URL to read
       if (Verbosity > 0)
         fprintf(stderr, "htmldoc: Reading %s...\n", filename);
 
-      _htmlPPI = _htmlBrowserWidth / PageWidth * 72.0f;
+      _htmlPPI = 72.0f * _htmlBrowserWidth / (PageWidth - PageLeft - PageRight);
 
       strcpy(base, file_directory(filename));
 
@@ -2004,21 +2124,16 @@ read_file(const char *filename,		// I  - File/URL to read
 
       fclose(docfile);
 
-      if (file->child != NULL)
-      {
-        if (*document == NULL)
-          *document = file;
-        else
-        {
-          while ((*document)->next != NULL)
-            *document = (*document)->next;
-
-          (*document)->next = file;
-          file->prev        = *document;
-        }
-      }
+      if (*document == NULL)
+        *document = file;
       else
-        htmlDeleteTree(file);
+      {
+        while ((*document)->next != NULL)
+          *document = (*document)->next;
+
+        (*document)->next = file;
+        file->prev        = *document;
+      }
     }
     else
     {
@@ -2060,7 +2175,7 @@ term_handler(int signum)	// I - Signal number
 static void
 usage(void)
 {
-  puts("HTMLDOC Version " SVERSION " Copyright 1997-2001 Easy Software Products, All Rights Reserved.");
+  puts("HTMLDOC Version " SVERSION " Copyright 1997-2002 Easy Software Products, All Rights Reserved.");
   puts("This software is governed by the GNU General Public License, Version 2, and");
   puts("is based in part on the work of the Independent JPEG Group.");
   puts("");
@@ -2079,12 +2194,13 @@ usage(void)
   puts("  --book");
   puts("  --bottom margin{in,cm,mm}");
   puts("  --browserwidth pixels");
-  puts("  --charset {8859-1...8859-15}");
+  puts("  --charset {cp-874...1258,iso-8859-1...8859-15,koi8-r}");
   puts("  --color");
   puts("  --compression[=level]");
   puts("  --datadir directory");
   puts("  --duplex");
   puts("  --effectduration {0.1..10.0}");
+  puts("  --embedfonts");
   puts("  --encryption");
   puts("  --firstpage {p1,toc,c1}");
   puts("  --fontsize {6.0..24.0}");
@@ -2112,15 +2228,17 @@ usage(void)
   puts("  --owner-password password");
   puts("  --no-compression");
   puts("  --no-duplex");
+  puts("  --no-embedfonts");
   puts("  --no-encryption");
   puts("  --no-links");
   puts("  --no-localfiles");
   puts("  --no-numbered");
   puts("  --no-pscommands");
+  puts("  --no-strict");
   puts("  --no-title");
   puts("  --no-toc");
-  puts("  --no-truetype");
   puts("  --numbered");
+  puts("  --nup {1,2,4,6,9,16}");
   puts("  {--outdir, -d} dirname");
   puts("  {--outfile, -f} filename.{ps,pdf,html}");
   puts("  --pageduration {1.0..60.0}");
@@ -2135,6 +2253,7 @@ usage(void)
   puts("  --quiet");
   puts("  --right margin{in,cm,mm}");
   puts("  --size {letter,a4,WxH{in,cm,mm},etc}");
+  puts("  --strict");
   puts("  --textcolor color");
   puts("  --textfont {courier,times,helvetica}");
   puts("  --title");
@@ -2145,9 +2264,9 @@ usage(void)
   puts("  --toclevels levels");
   puts("  --toctitle string");
   puts("  --top margin{in,cm,mm}");
-  puts("  --truetype");
   puts("  --user-password password");
   puts("  {--verbose, -v}");
+  puts("  --version");
   puts("  --webpage");
   puts("");
   puts("  fff = heading format string; each \'f\' can be one of:");
@@ -2174,5 +2293,5 @@ usage(void)
 
 
 /*
- * End of "$Id: htmldoc.cxx,v 1.36.2.35 2001/12/17 18:27:33 mike Exp $".
+ * End of "$Id: htmldoc.cxx,v 1.36.2.55 2002/10/25 13:37:54 mike Exp $".
  */
