@@ -1,9 +1,9 @@
 /*
- * "$Id: htmllib.cxx,v 1.41.2.44 2001/12/17 18:27:36 mike Exp $"
+ * "$Id: htmllib.cxx,v 1.41.2.64 2002/10/11 14:23:28 mike Exp $"
  *
  *   HTML parsing routines for HTMLDOC, a HTML document processing program.
  *
- *   Copyright 1997-2001 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -30,7 +30,6 @@
  *   htmlDeleteTree()    - Free all memory associated with a tree...
  *   htmlInsertTree()    - Insert a tree node to the parent.
  *   htmlNewTree()       - Create a new tree node for the parent.
- *   get_text()          - Get all text from the given tree.
  *   htmlGetText()       - Get all text from the given tree.
  *   htmlGetMeta()       - Get document "meta" data...
  *   htmlGetStyle()      - Get a style value from a node's STYLE attribute.
@@ -50,6 +49,8 @@
  *   compute_color()     - Compute the red, green, blue color from the given
  *   get_alignment()     - Get horizontal & vertical alignment values.
  *   fix_filename()      - Fix a filename to be relative to the base directory.
+ *   html_memory_used()  - Figure out the amount of memory that was used.
+ *   htmlDebugStats()    - Display debug statistics for HTML tree memory use.
  */
 
 /*
@@ -107,6 +108,15 @@ const char	*_htmlMarkups[] =
 		  "H4",
 		  "H5",
 		  "H6",
+		  "H7",
+		  "H8",
+		  "H9",
+		  "H10",
+		  "H11",
+		  "H12",
+		  "H13",
+		  "H14",
+		  "H15",
 		  "HEAD",
 		  "HR",
 		  "HTML",
@@ -170,7 +180,8 @@ typeface_t	_htmlBodyFont = TYPE_TIMES,
 		_htmlHeadingFont = TYPE_HELVETICA;
 
 int		_htmlInitialized = 0;	/* Initialized glyphs yet? */
-char		_htmlCharSet[256] = "";	/* Character set name */
+char		_htmlCharSet[256] = "iso-8859-1";
+					/* Character set name */
 float		_htmlWidths[4][4][256];	/* Character widths of fonts */
 const char	*_htmlGlyphsAll[65536];	/* Character glyphs for Unicode */
 const char	*_htmlGlyphs[256];	/* Character glyphs for charset */
@@ -216,8 +227,8 @@ static int	compare_variables(var_t *v0, var_t *v1);
 static int	compare_markups(uchar **m0, uchar **m1);
 static void	delete_node(tree_t *t);
 static void	insert_space(tree_t *parent, tree_t *t);
-static int	parse_markup(tree_t *t, FILE *fp);
-static int	parse_variable(tree_t *t, FILE *fp);
+static int	parse_markup(tree_t *t, FILE *fp, int *linenum);
+static int	parse_variable(tree_t *t, FILE *fp, int *linenum);
 static int	compute_size(tree_t *t);
 static int	compute_color(tree_t *t, uchar *color);
 static int	get_alignment(tree_t *t);
@@ -260,22 +271,27 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 		*t,		/* New tree entry */
 		*prev,		/* Previous tree entry */
 		*temp;		/* Temporary looping var */
-  int		pos;		/* Current file position */
+  int		descend;	/* Descend into node? */
   FILE		*embed;		/* File pointer for EMBED */
   char		newbase[1024];	/* New base directory for EMBED */
   uchar		*filename,	/* Filename for EMBED tag */
 		*face,		/* Typeface for FONT tag */
 		*color,		/* Color for FONT tag */
-		*size;		/* Size for FONT tag */
+		*size,		/* Size for FONT tag */
+		*type;		/* Type for EMBED tag */
   int		sizeval;	/* Size value from FONT tag */
+  int		linenum;	/* Line number in file */
   static uchar	s[10240];	/* String from file */
   static int	have_whitespace = 0;
   				/* Non-zero if there was leading whitespace */
 
 
+  DEBUG_printf(("htmlReadFile(parent=%p, fp=%p, base=\"%s\")\n",
+                parent, fp, base ? base : "(null)"));
+
 #ifdef DEBUG
-  strcat((char *)indent, "    ");
-#endif /* DEBUG */
+  indent[0] = '\0';
+#endif // DEBUG
 
  /*
   * Start off with no previous tree entry...
@@ -288,6 +304,8 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
   * Parse data until we hit end-of-file...
   */
 
+  linenum = 1;
+
   while ((ch = getc(fp)) != EOF)
   {
    /*
@@ -298,6 +316,9 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
     {
       while (isspace(ch))
       {
+	if (ch == '\n')
+	  linenum ++;
+
         have_whitespace = 1;
         ch              = getc(fp);
       }
@@ -362,9 +383,8 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       * Markup char; grab the next char to see if this is a /...
       */
 
-      pos = ftell(fp) - 1;
-
       ch = getc(fp);
+
       if (isspace(ch) || ch == '=' || ch == '<')
       {
        /*
@@ -373,6 +393,11 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	* be supported that we have added this hack...
 	*/
 
+	progress_error(HD_ERROR_HTML_ERROR, "Unquoted < on line %d.", linenum);
+
+	if (ch == '\n')
+	  linenum ++;
+
 	ptr = s;
 	if (have_whitespace)
 	{
@@ -402,12 +427,12 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	if (ch != '/')
           ungetc(ch, fp);
 
-	if (parse_markup(t, fp) == MARKUP_ERROR)
+	if (parse_markup(t, fp, &linenum) == MARKUP_ERROR)
 	{
-  #ifndef DEBUG
+#ifndef DEBUG
           progress_error(HD_ERROR_READ_ERROR,
-                         "Unable to parse HTML element at %d!", pos);
-  #endif /* !DEBUG */
+                         "Unable to parse HTML element on line %d!", linenum);
+#endif /* !DEBUG */
 
           delete_node(t);
           break;
@@ -443,6 +468,21 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
               break;
 	    }
 	}
+	else if (t->markup == MARKUP_BODY || t->markup == MARKUP_HEAD)
+	{
+	 /*
+          * Make sure we aren't inside an existing HEAD or BODY...
+	  */
+
+          for (temp = parent; temp != NULL; temp = temp->parent)
+            if (temp->markup == MARKUP_BODY || temp->markup == MARKUP_HEAD)
+              break;
+	    else if (temp->markup == MARKUP_EMBED)
+	    {
+	      temp = NULL;
+	      break;
+	    }
+	}
 	else if (t->markup == MARKUP_EMBED)
 	{
 	 /*
@@ -479,12 +519,11 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	      temp = NULL;
               break;
 	    }
-
 	}
 	else if (islentry(t->markup))
 	{
           for (temp = parent; temp != NULL; temp = temp->parent)
-            if (islentry(temp->markup) || isblock(temp->markup))
+            if (islentry(temp->markup))
               break;
 	    else if (islist(temp->markup) || issuper(temp->markup) ||
 	             istentry(temp->markup) || temp->markup == MARKUP_EMBED)
@@ -496,9 +535,10 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	else if (isblock(t->markup))
 	{
           for (temp = parent; temp != NULL; temp = temp->parent)
-            if (isblock(temp->markup) || islentry(temp->markup))
+            if (isblock(temp->markup))
               break;
 	    else if (istentry(temp->markup) || islist(temp->markup) ||
+	             islentry(temp->markup) ||
 	             issuper(temp->markup) || temp->markup == MARKUP_EMBED)
 	    {
 	      temp = NULL;
@@ -535,21 +575,109 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	{
           DEBUG_printf(("%s>>>> Auto-ascend <<<\n", indent));
 
-          delete_node(t);
+          if (temp && ch != '/' &&
+	      temp->markup != MARKUP_BODY &&
+	      temp->markup != MARKUP_DD &&
+	      temp->markup != MARKUP_DT &&
+	      temp->markup != MARKUP_HEAD &&
+	      temp->markup != MARKUP_HTML &&
+	      temp->markup != MARKUP_LI &&
+	      temp->markup != MARKUP_OPTION &&
+	      temp->markup != MARKUP_P &&
+	      temp->markup != MARKUP_TBODY &&
+	      temp->markup != MARKUP_TD &&
+	      temp->markup != MARKUP_TFOOT &&
+	      temp->markup != MARKUP_TH &&
+	      temp->markup != MARKUP_THEAD &&
+	      temp->markup != MARKUP_TR)
+	  {
+	    // Log this condition as an error...
+	    progress_error(HD_ERROR_HTML_ERROR,
+	                   "No /%s element before %s element on line %d.",
+	                   _htmlMarkups[temp->markup],
+			   _htmlMarkups[t->markup], linenum);
+	    DEBUG_printf(("%sNo /%s element before %s element on line %d.\n",
+	                  indent, _htmlMarkups[temp->markup],
+			  _htmlMarkups[t->markup], linenum));
+	  }
 
-	 /*
-          * If the markup doesn't start with a slash, or if it does but
-          * doesn't match up with the parent (i.e. <UL><LI>...<LI>...</UL>)
-          * then seek back so the parent entry gets a copy...
-          */
+#ifdef DEBUG
+          for (tree_t *p = parent;
+	       p && p != temp && indent[0];
+	       p = p->parent)
+	    indent[strlen((char *)indent) - 4] = '\0';
 
-          if (ch != '/' || temp != parent)
-            fseek(fp, pos, SEEK_SET);	/* Make sure parent gets this markup... */
-          break;
+          if (indent[0])
+            indent[strlen((char *)indent) - 4] = '\0';
+#endif // DEBUG
+
+
+          // Safety check; should never happen, since MARKUP_FILE is
+	  // the root node created by the caller...
+          if (temp->parent)
+	  {
+	    parent = temp->parent;
+            prev   = parent->last_child;
+	  }
+	  else
+	  {
+	    for (prev = temp; prev->next; prev = prev->next);
+	    parent = NULL;
+	  }
+
+          if (ch == '/')
+	  {
+	    // Closing element, so delete this node...
+            delete_node(t);
+	    continue;
+	  }
+	  else
+	  {
+	    // Reparent the node...
+	    if (parent == NULL)
+	    {
+	      t->halignment   = ALIGN_LEFT;
+	      t->valignment   = ALIGN_MIDDLE;
+	      t->typeface     = _htmlBodyFont;
+	      t->size         = SIZE_P;
+
+	      compute_color(t, _htmlTextColor);
+	    }
+	    else
+	    {
+	      t->link          = parent->link;
+	      t->halignment    = parent->halignment;
+	      t->valignment    = parent->valignment;
+	      t->typeface      = parent->typeface;
+	      t->size          = parent->size;
+	      t->style         = parent->style;
+	      t->superscript   = parent->superscript;
+	      t->subscript     = parent->subscript;
+	      t->preformatted  = parent->preformatted;
+	      t->indent        = parent->indent;
+	      t->red           = parent->red;
+	      t->green         = parent->green;
+	      t->blue          = parent->blue;
+	      t->underline     = parent->underline;
+	      t->strikethrough = parent->strikethrough;
+	    }
+	    
+          }
 	}
 	else if (ch == '/')
 	{
-          delete_node(t);
+	  // Log this condition as an error...
+	  if (t->markup != MARKUP_UNKNOWN &&
+	      t->markup != MARKUP_COMMENT)
+	  {
+	    progress_error(HD_ERROR_HTML_ERROR,
+	                   "Dangling /%s element on line %d.",
+			   _htmlMarkups[t->markup], linenum);
+	    DEBUG_printf(("%sDangling /%s element on line %d.\n",
+			  indent, _htmlMarkups[t->markup], linenum));
+          }
+
+	  delete_node(t);
 	  continue;
 	}
       }
@@ -569,19 +697,36 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
                (ch = getc(fp)) != EOF && (glyphptr - glyph) < 15;
                glyphptr ++)
             if (ch == ';' || isspace(ch))
+	    {
+	      if (ch == '\n')
+	        linenum ++;
+
               break;
+	    }
             else
               *glyphptr = ch;
 
-          *glyphptr = '\0';
-          ch = iso8859(glyph);
+          if (glyphptr == glyph)
+	  {
+	    progress_error(HD_ERROR_HTML_ERROR, "Unquoted & on line %d.",
+	                   linenum);
+            ch = '&';
+	  }
+	  else
+	  {
+            *glyphptr = '\0';
+            ch = iso8859(glyph);
+	  }
         }
 
         if (ch != 0 && ch != '\r')
           *ptr++ = ch;
 
         if (ch == '\n')
+	{
+	  linenum ++;
           break;
+	}
 
         ch = getc(fp);
       }
@@ -594,7 +739,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       t->markup = MARKUP_NONE;
       t->data   = (uchar *)strdup((char *)s);
 
-      DEBUG_printf(("%sfragment \"%s\"\n", indent, s));
+      DEBUG_printf(("%sfragment \"%s\", line %d\n", indent, s, linenum));
     }
     else
     {
@@ -621,8 +766,17 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
             else
               *glyphptr = ch;
 
-          *glyphptr = '\0';
-          ch = iso8859(glyph);
+          if (glyphptr == glyph)
+	  {
+	    progress_error(HD_ERROR_HTML_ERROR, "Unquoted & on line %d.",
+	                   linenum);
+            ch = '&';
+	  }
+	  else
+	  {
+            *glyphptr = '\0';
+            ch = iso8859(glyph);
+	  }
         }
 
         if (ch != 0)
@@ -631,6 +785,9 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
         ch = getc(fp);
       }
 
+      if (ch == '\n')
+	linenum ++;
+
       if (isspace(ch))
         have_whitespace = 1;
 
@@ -642,7 +799,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       t->markup = MARKUP_NONE;
       t->data   = (uchar *)strdup((char *)s);
 
-      DEBUG_printf(("%sfragment \"%s\"\n", indent, s));
+      DEBUG_printf(("%sfragment \"%s\", line %d\n", indent, s, linenum));
     }
 
    /*
@@ -664,7 +821,8 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
     t->prev   = prev;
     if (prev != NULL)
       prev->next = t;
-    else
+
+    if (tree == NULL)
       tree = t;
 
     prev = t;
@@ -673,6 +831,8 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
     * Do markup-specific stuff...
     */
 
+    descend = 0;
+
     switch (t->markup)
     {
       case MARKUP_BODY :
@@ -695,7 +855,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	                    (uchar *)fix_filename((char *)filename,
 			                          (char *)base));
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_IMG :
@@ -733,17 +893,35 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       case MARKUP_H4 :
       case MARKUP_H5 :
       case MARKUP_H6 :
+      case MARKUP_H7 :
+      case MARKUP_H8 :
+      case MARKUP_H9 :
+      case MARKUP_H10 :
+      case MARKUP_H11 :
+      case MARKUP_H12 :
+      case MARKUP_H13 :
+      case MARKUP_H14 :
+      case MARKUP_H15 :
           get_alignment(t);
 
           t->typeface      = _htmlHeadingFont;
-          t->size          = SIZE_H1 - t->markup + MARKUP_H1;
           t->subscript     = 0;
           t->superscript   = 0;
           t->strikethrough = 0;
           t->preformatted  = 0;
-          t->style         = STYLE_BOLD;
 
-          htmlReadFile(t, fp, base);
+	  if (t->markup > MARKUP_H6)
+          {
+	    t->size  = SIZE_H7;
+            t->style = STYLE_ITALIC;
+	  }
+	  else
+	  {
+            t->size  = SIZE_H1 - t->markup + MARKUP_H1;
+            t->style = STYLE_BOLD;
+	  }
+
+          descend = 1;
           break;
 
       case MARKUP_P :
@@ -757,7 +935,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
           t->strikethrough = 0;
           t->preformatted  = 0;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_PRE :
@@ -769,7 +947,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
           t->strikethrough = 0;
           t->preformatted  = 1;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_BLOCKQUOTE :
@@ -780,13 +958,13 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       case MARKUP_DL :
           t->indent ++;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_DIV :
           get_alignment(t);
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_HR :
@@ -802,9 +980,14 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       case MARKUP_LINK :
       case MARKUP_META :
       case MARKUP_WBR :
+      case MARKUP_COL :
           break;
 
       case MARKUP_EMBED :
+          if ((type = htmlGetVariable(t, (uchar *)"TYPE")) != NULL &&
+	      strncasecmp((const char *)type, "text/html", 9) != 0)
+	    break;
+
           if ((filename = htmlGetVariable(t, (uchar *)"SRC")) != NULL)
 	  {
 	    filename = (uchar *)fix_filename((char *)filename,
@@ -836,7 +1019,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 
           t->style = STYLE_BOLD;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_TD :
@@ -849,7 +1032,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 
           t->style = STYLE_NORMAL;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_FONT :
@@ -905,7 +1088,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
               t->size = sizeval;
           }
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_BIG :
@@ -922,7 +1105,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
           else
             t->size = 7;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_SMALL :
@@ -939,7 +1122,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
           else
             t->size = 0;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_SUP :
@@ -958,7 +1141,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	  else
 	    t->size = sizeval;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_SUB :
@@ -977,7 +1160,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 	  else
 	    t->size = sizeval;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_KBD :
@@ -1001,19 +1184,19 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 
           t->typeface = TYPE_COURIER;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_B :
           t->style = (style_t)(t->style | STYLE_BOLD);
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_DD :
           t->indent ++;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_VAR :
@@ -1021,7 +1204,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       case MARKUP_DFN :
           t->typeface = TYPE_HELVETICA;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_STRONG :
@@ -1032,7 +1215,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
       case MARKUP_I :
           t->style = (style_t)(t->style | STYLE_ITALIC);
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_U :
@@ -1047,7 +1230,7 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 
           t->underline = 1;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_STRIKE :
@@ -1063,13 +1246,13 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 
           t->strikethrough = 1;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       case MARKUP_CENTER :
           t->halignment = ALIGN_CENTER;
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
 
       default :
@@ -1079,14 +1262,20 @@ htmlReadFile(tree_t     *parent,/* I - Parent tree entry */
 
           get_alignment(t);
 
-          htmlReadFile(t, fp, base);
+          descend = 1;
           break;
     }
-  }  
 
+    if (descend)
+    {
 #ifdef DEBUG
-  indent[strlen((char *)indent) - 4] = '\0';
-#endif /* DEBUG */
+      strcat((char *)indent, "    ");
+#endif // DEBUG
+
+      parent = t;
+      prev   = NULL;
+    }
+  }  
 
   return (tree);
 }
@@ -1565,50 +1754,61 @@ htmlNewTree(tree_t   *parent,	/* I - Parent entry */
 }
 
 
-/*
- * 'get_text()' - Get all text from the given tree.
- */
-
-static uchar *		/* O - Pointer to last char set */
-get_text(tree_t *tree,	/* I - Tree to pick */
-         uchar  *buf)	/* I - Buffer to store text in */
-{
-  while (tree != NULL)
-  {
-    if (tree->child != NULL)
-      buf = get_text(tree->child, buf);
-    else if (tree->markup == MARKUP_NONE && tree->data != NULL)
-    {
-      strcpy((char *)buf, (char *)tree->data);
-      buf += strlen((char *)buf);
-    }
-    else if (tree->markup == MARKUP_BR)
-    {
-      strcat((char *)buf, " ");
-      buf ++;
-    }
-
-    tree = tree->next;
-  }
-
-  return (buf);
-}
-
-
 /*
  * 'htmlGetText()' - Get all text from the given tree.
  */
 
 uchar *				/* O - String containing text nodes */
-htmlGetText(tree_t *tree)	/* I - Tree to pick */
+htmlGetText(tree_t *t)		/* I - Tree to pick */
 {
-  uchar	buf[10240];		/* String buffer */
+  uchar		*s,		// String
+		*s2,		// New string
+		*tdata;		// Temporary string data
+  int		slen,		// Length of string
+		tlen;		// Length of node string
 
 
-  buf[0] = '\0';
-  get_text(tree, buf);
+  // Loop through all of the nodes in the tree and collect text...
+  slen = 0;
+  s    = NULL;
 
-  return ((uchar *)strdup((char *)buf));
+  while (t != NULL)
+  {
+    if (t->child)
+      tdata = htmlGetText(t->child);
+    else
+      tdata = t->data;
+
+    if (tdata != NULL)
+    {
+      // Add the text to this string...
+      tlen = strlen((char *)tdata);
+
+      if (slen)
+        s2 = (uchar *)realloc(s, 1 + slen + tlen);
+      else
+        s2 = (uchar *)malloc(1 + tlen);
+
+      if (!s2)
+        break;
+
+      s = s2;
+
+      memcpy((char *)s + slen, (char *)tdata, tlen);
+
+      slen += tlen;
+
+      if (t->child)
+	free(tdata);
+    }
+
+    t = t->next;
+  }
+
+  if (slen)
+    s[slen] = '\0';
+
+  return (s);
 }
 
 
@@ -1736,7 +1936,7 @@ htmlSetVariable(tree_t *t,	/* I - Tree entry */
 	key;			/* Search key */
 
 
-  DEBUG_printf(("%shtmlSetVariable(%08x, \"%s\", \"%s\")\n", indent, t, name,
+  DEBUG_printf(("%shtmlSetVariable(%p, \"%s\", \"%s\")\n", indent, t, name,
                 value ? (const char *)value : "(null)"));
 
   if (t->nvars == 0)
@@ -1945,13 +2145,13 @@ htmlSetCharSet(const char *cs)	/* I - Character set file to load */
       for (ch = 0; ch < 256; ch ++)
         _htmlWidths[i][j][ch] = 0.6f;
 
-      snprintf(filename, sizeof(filename), "%s/afm/%s", _htmlData,
+      snprintf(filename, sizeof(filename), "%s/fonts/%s.afm", _htmlData,
                _htmlFonts[i][j]);
       if ((fp = fopen(filename, "r")) == NULL)
       {
 #ifndef DEBUG
         progress_error(HD_ERROR_FILE_NOT_FOUND,
-                       "Unable to open font width file %s!", _htmlFonts[i][j]);
+                       "Unable to open font width file %s!", filename);
 #endif /* !DEBUG */
         continue;
       }
@@ -1973,10 +2173,7 @@ htmlSetCharSet(const char *cs)	/* I - Character set file to load */
 
           for (ch = 0; ch < 256; ch ++)
 	    if (_htmlGlyphs[ch] && strcmp(_htmlGlyphs[ch], glyph) == 0)
-	      break;
-
-          if (ch < 256)
-	    _htmlWidths[i][j][ch] = width * 0.001f;
+	      _htmlWidths[i][j][ch] = width * 0.001f;
 	}
 	else
 	{
@@ -1987,12 +2184,16 @@ htmlSetCharSet(const char *cs)	/* I - Character set file to load */
           if (sscanf(line, "%*s%d%*s%*s%f", &ch, &width) != 2)
 	    continue;
 
-          if (ch < 256)
+          if (ch < 256 && ch >= 0)
 	    _htmlWidths[i][j][ch] = width * 0.001f;
 	}
       }
 
       fclose(fp);
+
+      // Make sure that non-breaking space has the same width as
+      // a breaking space...
+      _htmlWidths[i][j][160] = _htmlWidths[i][j][32];
     }
 }
 
@@ -2029,7 +2230,11 @@ static int			/* O - -1 if m0 < m1, 0 if m0 == m1, 1 if m0 > m1 */
 compare_markups(uchar **m0,	/* I - First markup */
                 uchar **m1)	/* I - Second markup */
 {
-  return (strcasecmp((char *)*m0, (char *)*m1));
+  if (tolower((*m0)[0]) == 'h' && isdigit((*m0)[1]) &&
+      tolower((*m1)[0]) == 'h' && isdigit((*m1)[1]))
+    return (atoi((char *)*m0 + 1) - atoi((char *)*m1 + 1));
+  else
+    return (strcasecmp((char *)*m0, (char *)*m1));
 }
 
 
@@ -2126,7 +2331,8 @@ insert_space(tree_t *parent,	// I - Parent node
 
 static int			/* O - -1 on error, MARKUP_nnnn otherwise */
 parse_markup(tree_t *t,		/* I - Current tree entry */
-             FILE   *fp)	/* I - Input file */
+             FILE   *fp,	/* I - Input file */
+	     int    *linenum)	/* O - Current line number */
 {
   int	ch, ch2;		/* Characters from file */
   uchar	markup[255],		/* Markup string... */
@@ -2170,7 +2376,7 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
     * Unrecognized markup stuff...
     */
 
-    t->markup = MARKUP_COMMENT;
+    t->markup = MARKUP_UNKNOWN;
     strcpy((char *)comment, (char *)markup);
     cptr = comment + strlen((char *)comment);
 
@@ -2181,10 +2387,10 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
     t->markup = (markup_t)((const char **)temp - _htmlMarkups);
     cptr      = comment;
 
-    DEBUG_printf(("%s%s\n", indent, markup));
+    DEBUG_printf(("%s%s, line %d\n", indent, markup, *linenum));
   }
 
-  if (t->markup == MARKUP_COMMENT)
+  if (t->markup == MARKUP_COMMENT || t->markup == MARKUP_UNKNOWN)
   {
     while (ch != EOF && cptr < (comment + sizeof(comment) - 1))
     {
@@ -2193,6 +2399,9 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
 
       *cptr++ = ch;
 
+      if (ch == '\n')
+        (*linenum) ++;
+
       if (ch == '-')
       {
         if ((ch2 = getc(fp)) == '>')
@@ -2218,10 +2427,13 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
   {
     while (ch != EOF && ch != '>')
     {
+      if (ch == '\n')
+        (*linenum) ++;
+
       if (!isspace(ch))
       {
         ungetc(ch, fp);
-        parse_variable(t, fp);
+        parse_variable(t, fp, linenum);
       }
 
       ch = getc(fp);
@@ -2238,7 +2450,8 @@ parse_markup(tree_t *t,		/* I - Current tree entry */
 
 static int			/* O - -1 on error, 0 on success */
 parse_variable(tree_t *t,	/* I - Current tree entry */
-               FILE   *fp)	/* I - Input file */
+               FILE   *fp,	/* I - Input file */
+	       int    *linenum)	/* I - Current line number */
 {
   uchar	name[1024],		/* Name of variable */
 	value[10240],		/* Value of variable */
@@ -2255,9 +2468,17 @@ parse_variable(tree_t *t,	/* I - Current tree entry */
 
   *ptr = '\0';
 
+  if (ch == '\n')
+    (*linenum) ++;
+
   while (isspace(ch) || ch == '\r')
+  {
     ch = getc(fp);
 
+    if (ch == '\n')
+      (*linenum) ++;
+  }
+
   switch (ch)
   {
     default :
@@ -2283,6 +2504,8 @@ parse_variable(tree_t *t,	/* I - Current tree entry */
             else if (ptr < (value + sizeof(value) - 1) &&
 	             ch != '\n' && ch != '\r')
               *ptr++ = ch;
+	    else if (ch == '\n')
+	      (*linenum) ++;
 
           *ptr = '\0';
         }
@@ -2294,6 +2517,8 @@ parse_variable(tree_t *t,	/* I - Current tree entry */
             else if (ptr < (value + sizeof(value) - 1) &&
 	             ch != '\n' && ch != '\r')
               *ptr++ = ch;
+	    else if (ch == '\n')
+	      (*linenum) ++;
 
           *ptr = '\0';
         }
@@ -2306,6 +2531,9 @@ parse_variable(tree_t *t,	/* I - Current tree entry */
             else if (ptr < (value + sizeof(value) - 1))
               *ptr++ = ch;
 
+	  if (ch == '\n')
+	    (*linenum) ++;
+
           *ptr = '\0';
           if (ch == '>')
             ungetc(ch, fp);
@@ -2441,14 +2669,14 @@ compute_size(tree_t *t)		/* I - Tree entry */
       else if (*ptr == '\t')
         width = (float)(((int)width + 7) & ~7);
       else
-        width += _htmlWidths[t->typeface][t->style][*ptr];
+        width += _htmlWidths[t->typeface][t->style][(int)*ptr & 255];
 
    if (width < max_width)
      width = max_width;
   }
   else if (t->data)
     for (width = 0.0, ptr = t->data; *ptr != '\0'; ptr ++)
-      width += _htmlWidths[t->typeface][t->style][*ptr];
+      width += _htmlWidths[t->typeface][t->style][(int)*ptr & 255];
   else
     width = 0.0f;
 
@@ -2538,7 +2766,7 @@ get_alignment(tree_t *t)	/* I - Tree entry */
  * 'fix_filename()' - Fix a filename to be relative to the base directory.
  */
 
-static const char *				/* O - Fixed filename */
+static const char *			/* O - Fixed filename */
 fix_filename(char *filename,		/* I - Original filename */
              char *base)		/* I - Base directory */
 {
@@ -2571,10 +2799,10 @@ fix_filename(char *filename,		/* I - Original filename */
       strncmp(filename, ".\\", 2) == 0)
     filename += 2;
 
-  if (strncmp(base, "http://", 7) == 0)
+  if (strncmp(base, "http://", 7) == 0 || strncmp(base, "https://", 8) == 0)
   {
     strcpy(newfilename, base);
-    base = newfilename + 7;
+    base = strchr(newfilename, ':') + 3;
 
     if (filename[0] == '/')
     {
@@ -2645,14 +2873,77 @@ fix_filename(char *filename,		/* I - Original filename */
 #ifdef MAC
   strcat(newfilename, ":");
 #else
-  strcat(newfilename, "/");
+  if (filename[0] != '/' && *base && base[strlen(base) - 1] != '/')
+    strcat(newfilename, "/");
 #endif // MAC
+
   strcat(newfilename, filename);
 
   return (file_find(Path, newfilename));
 }
 
 
+//
+// 'html_memory_used()' - Figure out the amount of memory that was used.
+//
+
+static int				// O - Bytes used
+html_memory_used(tree_t *t)		// I - Tree node
+{
+  int	i;				// Looping var
+  int	bytes;				// Bytes used
+
+
+  if (t == NULL)
+    return (0);
+
+  bytes = 0;
+
+  while (t != NULL)
+  {
+    bytes += sizeof(tree_t);
+    bytes += t->nvars * sizeof(var_t);
+
+    for (i = 0; i < t->nvars; i ++)
+    {
+      bytes += (strlen((char *)t->vars[i].name) + 8) & ~7;
+
+      if (t->vars[i].value != NULL)
+        bytes += (strlen((char *)t->vars[i].value) + 8) & ~7;
+    }
+
+    if (t->data != NULL)
+      bytes += (strlen((char *)t->data) + 8) & ~7;
+
+    bytes += html_memory_used(t->child);
+
+    t = t->next;
+  }
+
+  return (bytes);
+}
+
+
+//
+// 'htmlDebugStats()' - Display debug statistics for HTML tree memory use.
+//
+
+void
+htmlDebugStats(const char *title,	// I - Title
+               tree_t     *t)		// I - Document root node
+{
+  const char	*debug;			/* HTMLDOC_DEBUG env var */
+
+
+  if ((debug = getenv("HTMLDOC_DEBUG")) == NULL ||
+      (strstr(debug, "all") == NULL && strstr(debug, "memory") == NULL))
+    return;
+
+  progress_error(HD_ERROR_NONE, "DEBUG: %s = %d kbytes", title,
+                 (html_memory_used(t) + 1023) / 1024);
+}
+
+
 /*
- * End of "$Id: htmllib.cxx,v 1.41.2.44 2001/12/17 18:27:36 mike Exp $".
+ * End of "$Id: htmllib.cxx,v 1.41.2.64 2002/10/11 14:23:28 mike Exp $".
  */
