@@ -1,9 +1,9 @@
 /*
- * "$Id: http.c,v 1.1.2.10 2001/10/20 21:49:21 mike Exp $"
+ * "$Id: http.c,v 1.1.2.14 2002/09/11 19:24:39 swdev Exp $"
  *
- *   HTTP routines for the Common UNIX Printing System (CUPS) scheduler.
+ *   HTTP routines for the HTMLDOC software.
  *
- *   Copyright 1997-2001 by Easy Software Products, all rights reserved.
+ *   Copyright 1997-2002 by Easy Software Products, all rights reserved.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -12,14 +12,14 @@
  *   file is missing or damaged please contact Easy Software Products
  *   at:
  *
- *       Attn: CUPS Licensing Information
+ *       Attn: HTMLDOC Licensing Information
  *       Easy Software Products
  *       44141 Airport View Drive, Suite 204
  *       Hollywood, Maryland 20636-3111 USA
  *
  *       Voice: (301) 373-9603
- *       EMail: cups-info@cups.org
- *         WWW: http://www.cups.org
+ *       EMail: info@easysw.com
+ *         WWW: http://www.easysw.com
  *
  * Contents:
  *
@@ -599,10 +599,11 @@ httpSeparate(const char *uri,		/* I - Universal Resource Identifier */
 	     int        *port,		/* O - Port number to use */
              char       *resource)	/* O - Resource/filename [1024] */
 {
-  char		*ptr;				/* Pointer into string... */
-  const char	*atsign,			/* @ sign */
-		*slash;				/* Separator */
-  char		safeuri[HTTP_MAX_URI];		/* "Safe" local copy of URI */
+  char		*ptr;			/* Pointer into string... */
+  const char	*atsign,		/* @ sign */
+		*slash;			/* Separator */
+  char		safeuri[HTTP_MAX_URI];	/* "Safe" local copy of URI */
+  char		quoted;			/* Quoted character */
 
 
  /*
@@ -726,7 +727,30 @@ httpSeparate(const char *uri,		/* I - Universal Resource Identifier */
 
     for (ptr = username; uri < atsign; uri ++)
       if (ptr < (username + HTTP_MAX_URI - 1))
-	*ptr++ = *uri;
+      {
+        if (*uri == '%' && isxdigit(uri[1]) && isxdigit(uri[2]))
+	{
+	 /*
+	  * Grab a hex-encoded username and password...
+	  */
+
+          uri ++;
+	  if (isalpha(*uri))
+	    quoted = (tolower(*uri) - 'a' + 10) << 4;
+	  else
+	    quoted = (*uri - '0') << 4;
+
+          uri ++;
+	  if (isalpha(*uri))
+	    quoted |= tolower(*uri) - 'a' + 10;
+	  else
+	    quoted |= *uri - '0';
+
+          *ptr++ = quoted;
+	}
+	else
+	  *ptr++ = *uri;
+      }
 
     *ptr = '\0';
 
@@ -2074,5 +2098,5 @@ http_upgrade(http_t *http)	/* I - HTTP data */
 
 
 /*
- * End of "$Id: http.c,v 1.1.2.10 2001/10/20 21:49:21 mike Exp $".
+ * End of "$Id: http.c,v 1.1.2.14 2002/09/11 19:24:39 swdev Exp $".
  */
