@@ -1,9 +1,9 @@
 /*
- * "$Id: util.cxx,v 1.1.2.9 2001/10/17 21:13:41 mike Exp $"
+ * "$Id: util.cxx,v 1.1.2.14 2002/08/12 18:08:36 mike Exp $"
  *
  *   Utility functions for HTMLDOC, a HTML document processing program.
  *
- *   Copyright 1997-2001 by Easy Software Products.
+ *   Copyright 1997-2002 by Easy Software Products.
  *
  *   These coded instructions, statements, and computer programs are the
  *   property of Easy Software Products and are protected by Federal
@@ -27,6 +27,7 @@
  *                       or letters.
  *   get_color()       - Get a standard color value...
  *   get_format()      - Convert an old "fff" format string to the new format.
+ *   get_fmt()         - Convert a new format string to the old "fff" format.
  *   get_measurement() - Get a size measurement in inches, points, centimeters,
  *                       or millimeters.
  *   set_page_size()   - Set the output page size.
@@ -196,6 +197,12 @@ get_color(const uchar *color,	/* I - Color attribute */
       rgb[1] = 0.0f;
       rgb[2] = 0.0f;
     }
+    else
+    {
+      rgb[0] = 1.0f;
+      rgb[1] = 1.0f;
+      rgb[2] = 1.0f;
+    }
     return;
   }
   else if (color[0] == '#')
@@ -216,7 +223,12 @@ get_color(const uchar *color,	/* I - Color attribute */
 	break;
 
     if (i >= (int)(sizeof(colors) / sizeof(colors[0])))
-      i = 1; /* Black */
+    {
+      if (defblack)
+        i = 1; /* Black */
+      else
+        i = 17; /* White */
+    }
 
     rgb[0] = colors[i].red / 255.0f;
     rgb[1] = colors[i].green / 255.0f;
@@ -229,7 +241,7 @@ get_color(const uchar *color,	/* I - Color attribute */
 // 'get_format()' - Convert an old "fff" format string to the new format.
 //
 
-extern void
+void
 get_format(const char *fmt,		// I - Old "fff" format
            char       **formats)	// O - New format strings
 {
@@ -237,6 +249,13 @@ get_format(const char *fmt,		// I - Old "fff" format
 
 
   for (i = 0; i < 3; i ++)
+  {
+    if (formats[i])
+    {
+      free(formats[i]);
+      formats[i] = NULL;
+    }
+
     switch (fmt[i])
     {
       case '/' :
@@ -303,14 +322,15 @@ get_format(const char *fmt,		// I - Old "fff" format
           formats[i] = NULL;
           break;
     }
+  }
 }
 
 
 //
-// 'get_format()' - Convert an old "fff" format string to the new format.
+// 'get_fmt()' - Convert a new format string to the old "fff" format.
 //
 
-extern const char *			// O - Old format string
+const char *				// O - Old format string
 get_fmt(char **formats)			// I - New format strings
 {
   int		i, j;			// Looping vars
@@ -417,6 +437,25 @@ set_page_size(const char *size)	/* I - Page size string */
     PageWidth  = 612;
     PageLength = 792;
   }
+  else if (strcasecmp(size, "legal") == 0)
+  {
+   /*
+    * US Legal - 8.5x14 inches (216x356mm).
+    */
+
+    PageWidth  = 612;
+    PageLength = 1008;
+  }
+  else if (strcasecmp(size, "tabloid") == 0 ||
+           strcasecmp(size, "b") == 0)
+  {
+   /*
+    * US Tabloid - 11x17 inches (279x432mm).
+    */
+
+    PageWidth  = 792;
+    PageLength = 1214;
+  }
   else if (strcasecmp(size, "a4") == 0)
   {
    /*
@@ -426,6 +465,15 @@ set_page_size(const char *size)	/* I - Page size string */
     PageWidth  = 595;
     PageLength = 842;
   }
+  else if (strcasecmp(size, "a3") == 0)
+  {
+   /*
+    * European standard A3 - 297x420mm (11.69x16.54 inches).
+    */
+
+    PageWidth  = 842;
+    PageLength = 1190;
+  }
   else if (strcasecmp(size, "universal") == 0)
   {
    /*
@@ -466,5 +514,5 @@ set_page_size(const char *size)	/* I - Page size string */
 
 
 /*
- * End of "$Id: util.cxx,v 1.1.2.9 2001/10/17 21:13:41 mike Exp $".
+ * End of "$Id: util.cxx,v 1.1.2.14 2002/08/12 18:08:36 mike Exp $".
  */
